import numpy as np
import pandas as pd
from scipy.stats import norm

from .IRR_result import IRR_result


def kappam_fleiss(ratings, exact=False, detail=False):
    """Computes Fleiss' Kappa as an index of interrater agreement between m raters on categorical data. Additionally,
    category-wise Kappas could be computed.

    Parameters
    ----------
    ratings: array_like
        subjects * raters array or dataframe
    exact: bool
        a boolean indicating whether the exact Kappa (Conger, 1980) or the Kappa described by Fleiss (1971) should be
        computed.
    detail: bool
        a boolean indicating whether category-wise Kappas should be computed

   """
    ratings = pd.DataFrame(ratings)
    ns = ratings.shape[0]
    nr = ratings.shape[1]

    lev = np.unique(ratings)

    ttab = ratings.apply(lambda row: pd.Categorical(row, lev).value_counts(), axis=1)

    ttab = ttab.values

    agreeP = np.sum((np.sum(ttab**2, axis=1) - nr) / (nr * (nr - 1)) / ns)

    if exact:
        method = "Fleiss` Kappa for m Raters (exact value)"
        rtab = ratings.apply(lambda col: pd.Categorical(col, lev).value_counts(), axis=0)
        rtab /= ns

        cov = np.apply_along_axis(np.cov, 1, rtab)
        chanceP = np.sum(np.sum(ttab, axis=0)**2) / (ns * nr)**2 - np.sum(cov * (nr - 1) / nr) / (nr - 1)
    else:
        method = "Fleiss` Kappa for m Raters"
        chanceP = np.sum(np.sum(ttab, axis=0)**2) / (ns * nr)**2

    value = (agreeP - chanceP) / (1 - chanceP)

    if not exact:
        pj = np.sum(ttab, axis=0) / (ns * nr)
        qj = 1 - pj

        varkappa = (2 / (sum(pj * qj)**2 * (ns * nr * (nr - 1)))) * (sum(pj * qj)**2 - sum(pj * qj * (qj - pj)))
        SEkappa = np.sqrt(varkappa)

        u = value/SEkappa
        pvalue = 2 * (1 - norm.cdf(np.abs(u)))

        if detail:
            pj = np.sum(ttab, axis=0) / (ns * nr)
            pjk = (np.sum(ttab**2, axis=0) - ns * nr * pj) / (ns * nr * (nr - 1) * pj)

            kappaK = (pjk - pj) / (1 - pj)

            varkappaK = 2 / (ns * nr * (nr-1))
            SEkappaK = np.sqrt(varkappaK)

            uK = kappaK / SEkappaK
            p_valueK = 2 * (1 - norm.cdf(np.abs(uK)))

            tableK = pd.DataFrame([kappaK, uK, p_valueK], columns=lev, index=["Kappa", "z", "p.value"])
            tableK = tableK.round(3).T

    rval = {"method": method, "subjects": ns, "raters": nr, "irr_name": "Kappa", "value": value}

    if not exact:
        if detail:
            rval = {**rval, "detail": tableK}

        rval = {**rval, "stat_name": "z", "statistic": u, "pvalue": pvalue}

    return IRR_result(**rval)



from math import ceil

from scipy.stats import norm


def N_cohen_kappa(rate1, rate2, k1, k0, alpha=0.05, power=0.8, twosided=False):
    """Sample Size Calculation for Cohen's Kappa Statistic.

    Parameters
    ----------
    rate1: float
        the probability that the first rater will record a positive diagnosis
    rate2: float
        the probability that the second rater will record a positive diagnosis
    k1: float
        the true Cohen's Kappa statistic
    k0: float
        the value of Kappa under the null hypothesis
    alpha: float
        type I error of data
    power: float
        the desired power to detect the difference between true Kappa and hypothetical Kappa
    twosided: bool
        True if test is two-sided
    """

    d = 2 if twosided else 1

    pi2 = 1 - rate1
    pi_2 = 1 - rate2

    pie = rate1 * rate2 + pi2 * pi_2
    pi0 = k1 * (1 - pie) + pie

    pi22 = (pi0 - rate1 + pi_2) / 2
    pi11 = pi0 - pi22
    pi12 = rate1 - pi11
    pi21 = rate2 - pi11

    pi0_h = k0 * (1 - pie) + pie

    pi22_h = (pi0_h - rate1 + pi_2) / 2
    pi11_h = pi0_h - pi22_h
    pi12_h = rate1 - pi11_h
    pi21_h = rate2 - pi11_h

    Q = (1 - pie)**-4 * (pi11 * (1 - pie - (rate2 + rate1) * ( 1 - pi0))**2 + pi22 * (1 - pie - (pi_2 + pi2) *
        (1 - pi0))**2 + (1 - pi0)**2 * (pi12 * (rate2 + pi2)**2 + pi21 * (pi_2 + rate1)**2) - (pi0 * pie - 2 *
        pie + pi0)**2)

    Q_h = (1 - pie)**-4 * (pi11_h * (1 - pie - (rate2 + rate1) * (1 - pi0_h))**2 + pi22_h * (1 - pie - (pi_2 + pi2) *
          (1 - pi0_h))**2 + (1 - pi0_h)**2 * (pi12_h * (rate2 + pi2)**2 + pi21_h * (pi2 + rate1)**2) - (pi0_h * pie -
          2 * pie + pi0_h)**2)

    N = ((norm.ppf(1 - alpha / d) * Q_h**0.5 + norm.ppf(power) * Q**0.5) / (k1 - k0))**2

    return ceil(N)



import numpy as np

from .IRR_result import IRR_result


def maxwell(ratings):
    """Calculate Maxwell's RE

    Parameters
    ----------
    ratings: array_like
        subjects * raters array or dataframe

    """
    ratings = np.array(ratings)  # make sure ratings is not a list or DataFrame

    ratings = ratings[~np.isnan(ratings).any(axis=1)]  # drop nans

    ns = ratings.shape[0]
    nr = ratings.shape[1]

    if nr > 2:
        raise Exception("Number of raters exceeds 2.")

    levels = set(ratings.ravel())  # number of unique levels

    if len(levels) > 2:
        raise Exception("Ratings are not binary.")

    r1, r2 = ratings[:, 0], ratings[:, 1]

    coeff = 2 * np.sum((r1 - r2) == 0) / ns - 1

    return IRR_result("Maxwell's RE", ns, nr, "RE", coeff)



import numpy as np
import pandas as pd
from scipy.stats import chi2

from .IRR_result import IRR_result


def kendall(ratings, correct=False):
    """Computes Kendall's coefficient of concordance as an index of interrater reliability of ordinal data. The
    coefficient could be corrected for ties within raters.

    Parameters
    ----------
    ratings: array_like
        subjects * raters array or dataframe
    correct: bool
        a logical indicating whether the coefficient should be corrected for ties within raters.

   """
    ratings = pd.DataFrame(ratings)
    ns = ratings.shape[0]
    nr = ratings.shape[1]

    ratings.dropna(inplace=True)  # drop nans

    if correct:
        ratings_rank = ratings.rank()
        error = None

        Tj = 0
        for col in ratings_rank:
            rater = ratings_rank[col].value_counts()
            ties = rater[rater > 1]
            Tj += np.sum(ties**3 - ties)

        coeff = (12 * np.cov(np.sum(ratings_rank, axis=1)) * (ns - 1)) / (nr**2 * (ns**3 - ns) - nr * Tj)
    else:
        testties = np.unique(ratings, axis=0)

        if not len(testties):
            error = "Coefficient may be incorrect due to ties"
        else:
            if len(testties) < len(ratings):
                error = "Coefficient may be incorrect due to ties"

        ratings_rank = ratings.rank()

        coeff = (12 * np.cov(np.sum(ratings_rank, axis=1)) * (ns - 1) / (nr**2 * (ns**3 - ns)))

    Xvalue = nr * (ns - 1) * coeff
    df1 = ns - 1
    pvalue = 1 - chi2.cdf(Xvalue, df1)

    method = "Kendall's coefficient of concordance"
    return IRR_result(method, ns, nr, "Wt", coeff, Xvalue, f"Chisq ({df1})", pvalue, error=error)



from .agree import agree
from .bhapkar import bhapkar
from .finn import finn
from .intraclass_correlation import intraclass_correlation
from .iota import iota
from .kappa2 import kappa2
from .kappam_fleiss import kappam_fleiss
from .kappam_light import kappam_light
from .kendall import kendall
from .kripp_alpha import kripp_alpha
from .maxwell import maxwell
from .meancor import meancor
from .meanrho import meanrho
from .N2_cohen_kappa import N2_cohen_kappa
from .N_cohen_kappa import N_cohen_kappa
from .rater_bias import rater_bias
from .read_data import read_data
from .rel_inter_intra import rel_inter_intra
from .robinson import robinson
from .stuart_maxwell_mh import stuart_maxwell_mh



import numpy as np
from scipy.stats import f

from .IRR_result import IRR_result


def finn(ratings, s_levels, model):
    """Computes the Finn coefficient as an index of the interrater reliability of quantitative data.

    Parameters
    ----------
    ratings: array_like
        subjects * raters array or dataframe
    s_levels: float
        the number of different rating categories
    model: bool
        a character string specifying if a '"oneway"' model (default) with row effects random, or a '"twoway"' model
        with column and row effects random should be applied.

    """
    ratings = np.array(ratings)  # make sure ratings is not a list or DataFrame

    ratings = ratings[~np.isnan(ratings).any(axis=1)]  # drop nans

    ns = ratings.shape[0]
    nr = ratings.shape[1]

    SS_total = np.cov(ratings.ravel()) * (ns * nr - 1)
    MSr = np.cov(np.mean(ratings, axis=1)) * nr
    MSw = np.sum(np.apply_along_axis(np.cov, axis=1, arr=ratings) / ns)
    MSc = np.apply_along_axis(np.cov, axis=0, arr=np.mean(ratings, axis=0)) * ns
    MSe = (SS_total - MSr * (ns - 1) - MSc * (nr - 1)) / ((ns - 1) * (nr - 1))

    MSexp = 1 / 12 * (s_levels**2 - 1)

    df1 = 10e100  # Inf, holistic approach
    df2 = ns * (nr - 1)

    if model == "oneway":
        coeff = 1 - (MSw / MSexp)
        Fvalue = MSexp / MSw
        pvalue = 1 - f.cdf(Fvalue, df1, df2)
    elif model == "twoway":
        coeff = 1 - (MSe / MSexp)
        Fvalue = MSexp / MSe
        pvalue = 1 - f.cdf(Fvalue, df1, df2)

    return IRR_result(f"Finn-Coefficient (Model={model})", ns, nr, "Finn", coeff, Fvalue, f"F(Inf,{df2})", pvalue)



import operator

import numpy as np
from scipy.stats import norm


def max_tau(Marg, Kp):
    Kappa = Kp
    Mrg = Marg
    Dim = len(Mrg)
    PIe = np.sum(Mrg * Mrg)
    PI0 = Kappa + PIe * (1 - Kappa)

    if PI0 <= 0:
        raise Exception("Invalid set of inputs")

    Part1 = PI0 * (1 - PIe) ** 2
    Part3 = (PI0 * PIe - 2 * PIe + PI0) ** 2
    m = np.zeros((Dim, Dim))

    for ii in range(Dim):
        for jj in range(Dim):
            if ii == jj:
                m[ii, jj] = -1 * (1 - PI0) * 2 * Mrg[ii] * (2 * (1 - PIe) - (1 - PI0) * 2 * Mrg[ii])
            else:
                m[ii, jj] = (1 - PI0) ** 2 * (Mrg[ii] + Mrg[jj]) ** 2

    AA = np.diag(np.diag(np.ones((Dim ** 2, Dim ** 2))))
    BB = np.zeros((Dim, Dim ** 2))
    CC = np.diag(np.diag(np.ones((Dim, Dim))))

    for i in range(Dim):
        BB[i, (i * Dim):((i + 1) * Dim)] = 1.
        if i < Dim - 1:
            CC = np.hstack([CC, np.diag(np.diag(np.ones((Dim, Dim))))])

    DD = np.ones(Dim ** 2)
    EE = np.diag(np.diag(np.ones((Dim, Dim)))).ravel()
    A = np.vstack([AA, BB, CC, DD, EE])

    f_con = A
    f_obj = m.ravel()
    f_dir = [operator.ge] * Dim**2 + [operator.eq] * (2 * (Dim + 1))
    f_rhs = np.hstack([np.zeros(Dim ** 2), np.tile(Mrg, 2), 1., PI0])

    Part2, _ = maximize_from_iterables(f_obj, f_con, f_dir, f_rhs)

    TauSq = Part1 + Part2 - Part3
    if TauSq <= 0:
        raise Exception("Invalid set of inputs")
    Tau = np.sqrt(TauSq) / (1 - PIe) ** 2
    return Tau


def maximize_from_iterables(objectives, constraints, directions, rh_constraints):
    try:
        from pulp import lpDot, lpSum, LpProblem, LpMaximize, LpVariable, value
    except ModuleNotFoundError:
        raise Exception("This function requires the pulp library, please install that library first.")

    prob = LpProblem("myProblem", LpMaximize)

    variable_list = []
    for i, objective in enumerate(objectives):
        variable_list.append(LpVariable("x" + str(i).zfill(3)))

    for constraint, direction, rh_constraint in zip(constraints, directions, rh_constraints):
        prob += direction(lpDot(constraint, variable_list), rh_constraint)

    prob += lpSum(lpDot(objectives, variable_list))
    prob.solve()

    result = 0
    values = []
    for v, m in zip(variable_list, objectives):
        result += value(v) * m
        values.append(v.varValue)
    return result, values


def N2_cohen_kappa(mrg, k1, k0, alpha=0.05, power=0.8, twosided=False):
    """Sample Size Calculation for Cohen's Kappa Statistic with more than one category

    Parameters
    ----------
    mrg: array-like
        a vector of marginal probabilities given by raters
    k1: float
        the true Cohen's Kappa statistic
    k0: float
        the value of Kappa under the null hypothesis
    alpha: float
        type I error of data
    power: float
        the desired power to detect the difference between true Kappa and hypothetical Kappa
    twosided: bool
        True if test is two-sided
    """

    mrg = np.asarray(mrg)

    if np.any(mrg < 0):
        raise Exception("At least one marginal probability is negative")
    if np.abs(k0) > 1:
        raise Exception("Invalid Value for Kappa under H0")
    if np.abs(k1) > 1:
        raise Exception("Invalid Value for Kappa under H1")
    if not np.abs(k1 - k0) > 0:
        raise Exception("Kappa under H0 must be different from H1")
    twosided = 2 if twosided else 1
    if len(mrg) >= 10:
        raise Exception("Valid only up to 10 categories")
    if len(mrg) <= 1:
        raise Exception("Valid only for greater than equal to 2 categories")
    if not np.sum(mrg) == 1:
        raise Exception("Sum of marginal probabilities is not 1")
    if alpha <= 0 or alpha >= 1:
        raise Exception("Alpha must be between 0 and 1")
    if power <= 0 or power >= 1:
        raise Exception("Power must be between 0 and 1")
    if not power-alpha > 0:
        raise Exception("Power is less than Alpha")

    tau_null = max_tau(mrg, k0)
    tau_alt = max_tau(mrg, k1)
    z_alpha = norm.ppf(1 - alpha / twosided)
    z_beta = norm.ppf(power)
    raw_n = (((z_alpha * tau_null) + (z_beta * tau_alt)) / (k0 - k1))**2
    n = np.ceil(raw_n)

    return int(n)



import numpy as np
import pandas as pd
from scipy.stats import chi2

from .IRR_result import IRR_result


def stuart_maxwell_mh(ratings):
    """Calculates the Stuart-Maxwell coefficient of concordance for two raters.

    Parameters
    ----------
    ratings: array_like
        subjects * raters array or dataframe

    """
    ratings = pd.DataFrame(ratings)

    smx = pd.crosstab(ratings.iloc[:, 0], ratings.iloc[:, 1])
    smx = np.array(smx)

    rowsums = np.sum(smx, axis=1)
    colsums = np.sum(smx, axis=0)
    equalsums = rowsums == colsums

    if np.any(equalsums):
        smx = smx[~equalsums, ~equalsums]
        if smx.shape[0] < 2:
            raise Exception("Too many equal marginals, cannot compute")
        rowsums = np.sum(smx, axis=1)
        colsums = np.sum(smx, axis=0)

    k_minus1 = len(rowsums) - 1
    smd = (rowsums - colsums)[:k_minus1]
    smS = np.zeros((k_minus1, k_minus1))

    for i in range(k_minus1):
        for j in range(k_minus1):
            if i == j:
                smS[i, j] = rowsums[i] + colsums[j] - 2 * smx[i, j]
            else:
                smS[i, j] = -(smx[i, j] + smx[j, i])

    smstat = smd.T @ np.linalg.inv(smS) @ smd

    p = 1 - chi2.cdf(smstat, k_minus1)

    return IRR_result("Stuart-Maxwell marginal homogeneity", int(np.sum(smx)), 2, "Chisq", smstat, smstat,
                      f"Chisq({k_minus1})", p)




import numpy as np
from scipy.stats import f
from dataclasses import dataclass, asdict


@dataclass
class ICC_result:
    subjects: int
    raters: int
    model: str
    mtype: str
    name: str
    value: float
    r0: float
    Fvalue: float
    df1: float
    df2: float
    pvalue: float
    conf_level: float
    lower_bound: float
    upper_bound: float

    def to_dict(self):
        return asdict(self)

    def __repr__(self):
        model_string = "=" * 50 + "\n" + "Intraclass Correlation Results".center(50, " ") + "\n" + "=" * 50 + "\n"
        model_string += f"Model: {self.model}\nType: {self.mtype}\n\n"
        model_string += f"Subjects = {self.subjects}\nRaters = {self.raters}\n{self.name} = {self.value:.2f}\n\n"
        model_string += f"F-Test, H0: r0 = {self.r0} ; H1 : r0 > {self.r0}\n"
        model_string += f"F({self.df1:.2f},{self.df2:.2f}) = {self.Fvalue:.2f}, p = {self.pvalue:.4f}\n\n"
        model_string += f"{self.conf_level*100:.0f}%-Confidence Interval for ICC Population Values:\n"
        model_string += f"{self.lower_bound:.3f} < ICC < {self.upper_bound:.3f}\n"
        model_string += "=" * 50
        return model_string


def intraclass_correlation(ratings, model="oneway", mtype="consistency", unit="single", r0=0, conf_level=0.95):
    """Calculate the intraclass correlation for a set of ratings

    Parameters
    ----------
    ratings: array_like
        subjects * raters array or dataframe
    model: {"oneway", "twoway"}
        whether it is a "oneway" or "twoway" model
    mtype: {"consistency", "agreement"}
        whether you want to test for "consistency" or "agreement"
    unit: {"single", "average"}
        something
    r0: int
        something
    conf_level: float
        confidence level

    """
    ratings = np.array(ratings)  # make sure ratings is not a list or DataFrame

    alpha = 1 - conf_level
    ratings = ratings[~np.isnan(ratings).any(axis=1)]  # drop nans

    ns = ratings.shape[0]
    nr = ratings.shape[1]

    SS_total = np.cov(np.ravel(ratings)) * (ns * nr - 1)
    MSr = np.cov(np.mean(ratings, axis=1)) * nr
    MSw = np.sum(np.apply_along_axis(np.cov, axis=1, arr=ratings) / ns)
    MSc = np.apply_along_axis(np.cov, axis=0, arr=np.mean(ratings, axis=0)) * ns
    MSe = (SS_total - MSr * (ns - 1) - MSc * (nr - 1)) / ((ns - 1) * (nr - 1))

    if unit == "single":
        if model == "oneway":  # Assendorpf & Wallbot, S. 245, ICu, Bartko (1966) [3]
            name = "ICC(1)"
            coeff = (MSr - MSw) / (MSr + (nr - 1) * MSw)
            Fvalue = MSr / MSw * ((1 - r0) / (1 + (nr - 1) * r0))
            df1 = ns - 1
            df2 = ns * (nr - 1)
            pvalue = 1 - f.cdf(Fvalue, df1, df2)

            # confindence interval
            FL = (MSr / MSw) / f.ppf(1 - alpha/2, ns - 1, ns * (nr - 1))
            FU = (MSr / MSw) * f.ppf(1 - alpha/2, ns * (nr - 1), ns - 1)
            lbound = (FL - 1) / (FL + nr - 1)
            ubound = (FU - 1) / (FU + nr - 1)
        elif model == "twoway":
            if mtype == "consistency":
                # Asendorpf & Wallbott, S. 245, ICa
                # Bartko (1966), [21]
                # Shrout & Fleiss (1979), ICC(3,1)
                name = "ICC(C,1)"
                coeff = (MSr - MSe) / (MSr + (nr - 1) * MSe)
                Fvalue = MSr / MSe * ((1 - r0) / (1 + (nr - 1) * r0))
                df1 = ns - 1
                df2 = (ns - 1) * (nr - 1)
                pvalue = 1 - f.cdf(Fvalue, df1, df2)

                # confindence interval
                FL = (MSr / MSe) / f.ppf(1 - alpha / 2, ns - 1, (ns - 1) * (nr - 1))
                FU = (MSr / MSe) * f.ppf(1 - alpha / 2, (ns - 1) * (nr - 1), ns - 1)
                lbound = (FL - 1) / (FL + (nr - 1))
                ubound = (FU - 1) / (FU + (nr - 1))
            elif mtype == "agreement":
                # Asendorpf & Wallbott, S. 246, ICa'
                # Bartko (1966), [15]
                # Shrout & Fleiss (1979), ICC(2,1)
                name = "ICC(A,1)"
                coeff = (MSr - MSe) / (MSr + (nr - 1) * MSe + (nr / ns) * (MSc - MSe))
                a = (nr * r0) / (ns * (1 - r0))
                b = 1 + (nr * r0 * (ns - 1)) / (ns * (1 - r0))
                Fvalue = MSr / (a * MSc + b * MSe)
                a = (nr * coeff) / (ns * (1 - coeff))
                b = 1 + (nr * coeff * (ns - 1)) / (ns * (1 - coeff))
                v = (a * MSc + b * MSe)**2 / ((a * MSc)**2 / (nr - 1) + (b * MSe)**2 / ((ns - 1) * (nr - 1)))
                df1 = ns - 1
                df2 = v
                pvalue = 1 - f.cdf(Fvalue, df1, df2)

                # confidence interval (McGraw & Wong, 1996)
                FL = f.ppf(1 - alpha / 2, ns - 1, v)
                FU = f.ppf(1 - alpha / 2, v, ns - 1)
                lbound = (ns * (MSr - FL * MSe)) / (FL * (nr * MSc + (nr * ns - nr - ns) * MSe) + ns * MSr)
                ubound = (ns * (FU * MSr - MSe)) / (nr * MSc + (nr * ns - nr - ns) * MSe + ns * FU * MSr)
    elif unit == "average":
        if model == "oneway":
            # Asendorpf & Wallbott, S. 245, Ru
            name = f"ICC({nr})"
            coeff = (MSr - MSw) / MSr
            Fvalue = MSr / MSw * (1 - r0)
            df1 = ns - 1
            df2 = ns * (nr - 1)
            pvalue = 1 - f.cdf(Fvalue, df1, df2)

            # confidence interval
            FL = (MSr / MSw) / f.ppf(1 - alpha / 2, ns - 1, ns * (nr - 1))
            FU = (MSr / MSw) * f.ppf(1 - alpha / 2, ns * (nr - 1), ns - 1)
            lbound = 1 - 1 / FL
            ubound = 1 - 1 / FU
        elif model == "twoway":
            if mtype == "consistency":
                # Asendorpf & Wallbott, S. 246, Ra
                name = f"ICC(C,{nr})"
                coeff = (MSr - MSe) / MSr
                Fvalue = MSr / MSe * (1 - r0)
                df1 = ns - 1
                df2 = (ns - 1) * (nr - 1)
                pvalue = 1 - f.cdf(Fvalue, df1, df2)

                # confidence interval
                FL = (MSr / MSe) / f.ppf(1 - alpha / 2, ns - 1, (ns - 1) * (nr - 1))
                FU = (MSr / MSe) * f.ppf(1 - alpha / 2, (ns - 1) * (nr - 1), ns - 1)
                lbound = 1 - 1 / FL
                ubound = 1 - 1 / FU
            elif mtype == "agreement":
                name = f"ICC(A,{nr})"
                coeff = (MSr - MSe) / (MSr + (MSc - MSe) / ns)
                a = r0 / (ns * (1 - r0))
                b = 1 + (r0 * (ns - 1)) / (ns * (1 - r0))
                v = (a * MSc + b * MSe)**2 / ((a * MSc)**2 / (nr - 1) + (b * MSe)**2 / ((ns - 1) * (nr - 1)))
                Fvalue = MSr / (a * MSc + b * MSe)
                df1 = ns - 1
                df2 = v
                pvalue = 1 - f.cdf(Fvalue, df1, df2)

                # confidence interval (McGraw & Wong, 1996)
                FL = f.ppf(1 - alpha / 2, ns - 1, v)
                FU = f.ppf(1 - alpha / 2, v, ns - 1)
                lbound = (ns * (MSr - FL * MSe)) / (FL * (MSc - MSe) + ns * MSr)
                ubound = (ns * (FU * MSr - MSe)) / (MSc - MSe + ns * FU * MSr)

    return ICC_result(ns, nr, model, mtype, name, coeff, r0, Fvalue, df1, df2, pvalue, conf_level, lbound, ubound)



import numpy as np
from scipy.stats import norm, pearsonr

from .IRR_result import IRR_result


def meancor(ratings, fisher=True):
    """Computes the mean of bivariate Pearson's product moment correlations between raters as an index of the interrater
    reliability of quantitative data.

    Parameters
    ----------
    ratings: array_like
        subjects * raters array or dataframe
    fisher: bool
        a boolean indicating whether the correlation coefficients should be Fisher z-standardized before averaging.

    """
    ratings = np.array(ratings)  # make sure ratings is not a list or DataFrame

    ratings = ratings[~np.isnan(ratings).any(axis=1)]  # drop nans

    ns = ratings.shape[0]
    nr = ratings.shape[1]

    r = []
    for i in range(nr-1):
        for j in range(i + 1, nr):
            r.append(pearsonr(ratings[:, i], ratings[:, j])[0])

    r = np.array(r)
    delr = 0

    if fisher:
        delr = len(r) - len(r[(r < 1) & (r > -1)])
        # Eliminate perfect correlations (r=1, r=-1)
        r = r[(r < 1) & (r > -1)]

        rz = 0.5 * np.log((1 + r) / (1 - r))
        mrz = np.mean(rz)

        coeff = (np.exp(2 * mrz) - 1) / (np.exp(2 * mrz) + 1)
        SE = np.sqrt(1 / (ns - 3))

        u = coeff / SE
        pvalue = 2 * (1 - norm.cdf(np.abs(u)))

    else:
        coeff = np.mean(r)

    result = {"method": "Mean of bivariate correlations R", "subjects": ns, "raters": nr, "irr_name": "R",
              "value": coeff}
    if fisher:
        result = {**result, "statistic": u, "stat_name": "z", "pvalue": pvalue}
    if delr > 0:
        error = "perfect correlation was" if delr == 1 else "perfect correlations were"
        result = {**result, "error": error + "dropped before averaging"}

    return IRR_result(**result)



import numpy as np

from .IRR_result import IRR_result


def robinson(ratings):
    """Computes Robinson’s A as an index of the interrater reliability of quantitative data.

    Parameters
    ----------
    ratings: array_like
        subjects * raters array or dataframe

    """
    ratings = np.asarray(ratings)  # make sure ratings is not a list or DataFrame

    ratings = ratings[~np.isnan(ratings).any(axis=1)]  # drop nans

    ns = ratings.shape[0]
    nr = ratings.shape[1]

    SS_total = np.cov(np.ravel(ratings)) * (ns * nr - 1)
    SSb = np.cov(ratings.mean(axis=1)) * nr * (ns - 1)
    SSw = np.cov(ratings.mean(axis=0)) * ns * (nr - 1)
    SSr = SS_total - SSb - SSw

    coeff = SSb / (SSb + SSr)

    return IRR_result("Robinson's A", ns, nr, "A", coeff)



import numpy as np

from .IRR_result import IRR_result


def coincidence_matrix(array):
    array = np.asarray(array)
    levx = np.unique(array)
    levx = levx[~np.isnan(levx)]
    nval = len(levx)
    cm = np.zeros((nval, nval))
    dimx = array.shape

    if np.any(np.isnan(array)):
        mc = np.sum(~np.isnan(array), axis=0) - 1
    else:
        mc = np.ones(dimx[1])

    for col in range(dimx[1]):  # not very Pythonic I know, I'm sorry Guido
        for i1 in range(dimx[0] - 1):
            for i2 in range(i1 + 1, dimx[0]):
                if not np.isnan(array[i1, col]) and not np.isnan(array[i2, col]):
                    index1 = np.nonzero(levx == array[i1, col])
                    index2 = np.nonzero(levx == array[i2, col])
                    cm[index1, index2] += (1 + (index1 == index2)) / mc[col]
                    if index1 != index2:
                        cm[index2, index1] = cm[index1, index2]
    nmv = np.sum(np.sum(cm, axis=0))
    return cm, nmv, levx


def kripp_alpha(ratings, method=None):
    """Calculates the alpha coefficient of reliability proposed by Krippendorff

    Parameters
    ----------
    ratings: array_like
        observation x rater array or dataframe
    method: {"nominal", "ordinal", "interval", "ratio"}
        data level of x

   """
    ratings = np.asarray(ratings).T
    cm, nmv, levx = coincidence_matrix(ratings)
    dimcm = cm.shape

    triu1, triu2 = np.triu_indices_from(cm, k=len(cm)//2 - 1)
    triu2inds = triu2.argsort()  # R sorts rows first, numpy columns first
    triu1 = triu1[triu2inds]
    triu2 = triu2[triu2inds]
    utcm = cm[triu1, triu2]

    nc = np.sum(cm, axis=1)
    dv = levx

    diff2 = np.zeros(len(utcm))
    ncnk = np.zeros(len(utcm))
    ck = 0

    if dimcm[1] < 2:
        value = 1.
    else:
        for k in range(1, dimcm[1]):
            for c in range(k):
                ncnk[ck] = nc[c] * nc[k]
                if method == "nominal":
                    diff2[ck] = 1.
                elif method == "ordinal":
                    diff2[ck] = nc[c] / 2
                    if k > c:
                        for g in range(c+1, k):
                            diff2[ck] += nc[g]
                        diff2[ck] += nc[k] / 2
                        diff2[ck] **= 2
                elif method == "interval":
                    diff2[ck] = (dv[c] - dv[k])**2
                elif method == "ratio":
                    diff2[ck] = (dv[c] - dv[k])**2 / (dv[c] + dv[k])**2
                else:
                    raise Exception("Please specify a nominal/ordinal/interval/ratio data level")

                ck += 1

        value = 1 - (nmv - 1) * np.sum(utcm * diff2) / np.sum(ncnk * diff2)

    return IRR_result(f"Krippendorff's alpha ({method})", ratings.shape[1], ratings.shape[0], "alpha", value)



import numpy as np
import pandas as pd

from .IRR_result import IRR_result


def agree(ratings, tolerance=0, numeric=True):
    """Calculate the percentage agreement among raters

    Parameters
    ----------
    ratings: array_like
        subjects * raters array or dataframe
    tolerance: float
        tolerance in ratings before raters differ in agreement
    numeric: bool
        should data be treated as numeric or as categories

    """
    ratings = pd.DataFrame(ratings)  # make sure ratings is not a list or DataFrame

    ratings.dropna(inplace=True)  # drop nans

    ns = ratings.shape[0]
    nr = ratings.shape[1]

    if numeric:
        range_tab = np.max(ratings, axis=1) - np.min(ratings, axis=1)
        coeff = 100 * np.sum(range_tab <= tolerance) / ns
    else:
        range_tab = ratings.apply(lambda row: len(row.value_counts()), axis=1)
        coeff = 100 * (np.sum(range_tab == 1) / ns)
        tolerance = 0

    return IRR_result(f"Percentage agreement (Tolerance={tolerance:.2f})", ns, nr, "%-agree", coeff)



import pandas as pd

from pkg_resources import resource_filename


def read_data(name):
    """Reads sample data into DataFrame

    Parameters
    ----------
    name: str
        {"anxiety", "diagnoses", "gonio", "nmm", "photo", "video", "vision"}

    """

    if name == "photo":
        return pd.read_pickle(resource_filename("pyirr", "data/photo.p"))
    else:
        return pd.read_csv(resource_filename("pyirr", f"data/{name}.csv"), index_col=0)
